@@ -7,7 +7,7 @@
 # ==============================================================================
 # MIT License
 #
-# Copyright (c) 2019 Albert Moky
+# Copyright (c) 2020 Albert Moky
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -38,9 +38,9 @@
 
 from typing import Optional
 
-from udp.mtp.data import UInt16Data, UInt32Data
-from udp.mtp.data import bytes_to_int, uint8_to_bytes, uint16_to_bytes, uint32_to_bytes
-from udp.tlv import TLV, Type, Length, Value
+from udp.tlv.data import bytes_to_int, uint8_to_bytes, uint16_to_bytes, uint32_to_bytes
+from udp.tlv import UInt16Data, UInt32Data
+from udp.tlv import Tag, Length, Value, TagLengthValue
 
 """
     STUN Attributes
@@ -65,7 +65,27 @@ from udp.tlv import TLV, Type, Length, Value
 
 
 class AttributeValue(Value):
-    pass
+
+    @classmethod
+    def parse(cls, data: bytes, tag: Tag, length: Length=None):
+        if cls is AttributeValue:
+            # get attribute parser with type
+            assert isinstance(tag, AttributeType), 'attribute type error: %s' % tag
+            parser = s_attribute_parsers.get(tag)
+            if parser is not None:
+                # create instance by subclass
+                return parser.parse(data=data, tag=tag, length=length)
+        # check length
+        if length is None or length.value == 0:
+            return None
+        else:
+            length = length.value
+            data_len = len(data)
+            if data_len < length:
+                return None
+            elif data_len > length:
+                data = data[:length]
+        return cls(data=data)
 
 
 class AttributeLength(UInt16Data, Length):
@@ -75,8 +95,15 @@ class AttributeLength(UInt16Data, Length):
             data = uint16_to_bytes(value)
         super().__init__(data=data, value=value)
 
+    @classmethod
+    def parse(cls, data: bytes, tag: Tag):
+        length = super().parse(data=data, tag=tag)
+        if length is not None:
+            assert length.value & 0x0003 == 0, 'attribute length error: %d' % length.value
+            return length
 
-class AttributeType(UInt16Data, Type):
+
+class AttributeType(UInt16Data, Tag):
 
     def __init__(self, value: int, data: bytes=None, name: str=None):
         if data is None:
@@ -148,45 +175,32 @@ AlternateServer = AttributeType(0x8023, name='ALTERNATE-SERVER')
 Fingerprint = AttributeType(0x8028, name='FINGERPRINT')
 
 
-class Attribute(TLV):
+class Attribute(TagLengthValue):
 
-    def __init__(self, t: AttributeType, v: AttributeValue, data: bytes=None):
+    def __init__(self, tag: AttributeType, value: AttributeValue, data: bytes=None):
         if data is None:
-            data = t.data + uint16_to_bytes(len(v.data)) + v.data
-        super().__init__(data=data, t=t, v=v)
+            data = tag.data + uint16_to_bytes(len(value.data)) + value.data
+        super().__init__(data=data, tag=tag, value=value)
 
     def __str__(self):
         clazz = self.__class__.__name__
-        return '/* %s */ %s: %s' % (clazz, self.type, self.value)
+        return '/* %s */ %s: %s' % (clazz, self.tag, self.value)
 
     def __repr__(self):
         clazz = self.__class__.__name__
-        return '/* %s */ %s: %s' % (clazz, self.type, self.value)
+        return '/* %s */ %s: %s' % (clazz, self.tag, self.value)
 
     @classmethod
-    def parse_type(cls, data: bytes) -> Optional[AttributeType]:
+    def parse_tag(cls, data: bytes) -> Optional[AttributeType]:
         return AttributeType.parse(data=data)
 
     @classmethod
-    def parse_length(cls, data: bytes, t: AttributeType) -> Optional[AttributeLength]:
-        length = AttributeLength.parse(data=data, t=t)
-        assert (length.value & 0x0003) == 0, 'attribute length error: %s' % data
-        return length
+    def parse_length(cls, data: bytes, tag: AttributeType) -> Optional[AttributeLength]:
+        return AttributeLength.parse(data=data, tag=tag)
 
     @classmethod
-    def parse_value(cls, data: bytes, t: AttributeType, length: AttributeLength = None) -> Optional[Value]:
-        if length is not None:
-            # check length
-            data_len = len(data)
-            if data_len > length.value:
-                data = data[:length.value]
-            else:
-                assert data_len == length.value, 'data length not enough: %d < %d' % (data_len, length.value)
-        # get attribute parser with type
-        parser = s_attribute_parsers.get(t)
-        if parser is None:
-            parser = AttributeValue
-        return parser.parse(data=data, t=t, length=length)
+    def parse_value(cls, data: bytes, tag: AttributeType, length: AttributeLength = None) -> Optional[Value]:
+        return AttributeValue.parse(data=data, tag=tag, length=length)
 
 
 """
@@ -268,10 +282,9 @@ class MappedAddressValue(AttributeValue):
             array = ip.split('.')
             assert len(array) == 4, 'IP address error: %s' % ip
             return bytes([int(x) for x in array])
-            pass
         elif family == cls.family_ipv6:
             # TODO: IPv6
-            assert False, 'implement me!'
+            assert False, 'IPv6 not support yet'
         else:
             raise ValueError('unknown address family: %d' % family)
 
@@ -290,13 +303,25 @@ class MappedAddressValue(AttributeValue):
         elif family == cls.family_ipv6:
             assert len(address) == 16, 'IPv6 data error: %s' % address
             # TODO: IPv6
-            assert False, 'implement me!'
+            assert False, 'IPv6 not support yet'
         else:
             raise ValueError('unknown address family: %d' % family)
 
     @classmethod
-    def parse(cls, data: bytes, t: Type, length: Length=None):
-        assert len(data) >= 8, 'mapped-address value error: %s' % data
+    def parse(cls, data: bytes, tag: Tag, length: Length=None):
+        # check length
+        if length is None or (length.value != 8 and length.value != 20):
+            # raise ValueError('mapped-address length error: %s' % length)
+            return None
+        else:
+            length = length.value
+            data_len = len(data)
+            if data_len < length:
+                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
+                return None
+            elif data_len > length:
+                data = data[:length]
+        # checking
         if data[0] != 0:
             return None
         family = bytes_to_int(data[1:2])
@@ -540,8 +565,20 @@ class ChangeRequestValue(UInt32Data, AttributeValue):
         return '%d' % self.value
 
     @classmethod
-    def parse(cls, data: bytes, t: Type, length: Length=None):
-        assert length.value == 4, 'Change-Request value error: %s' % length
+    def parse(cls, data: bytes, tag: Tag, length: Length=None):
+        # check length
+        if length is None or length.value != 4:
+            # raise ValueError('Change-Request value error: %s' % length)
+            return None
+        else:
+            length = length.value
+            data_len = len(data)
+            if data_len < length:
+                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
+                return None
+            elif data_len > length:
+                data = data[:length]
+        # get value
         value = bytes_to_int(data)
         if value == ChangeIPAndPort.value:
             return ChangeIPAndPort
@@ -602,16 +639,19 @@ class SoftwareValue(AttributeValue):
         return self.__desc
 
     @classmethod
-    def parse(cls, data: bytes, t: Type, length: Length=None):
+    def parse(cls, data: bytes, tag: Tag, length: Length=None):
+        # check length
         if length is None or length.value == 0:
             return None
         else:
             length = length.value
-        data_len = len(data)
-        if data_len < length:
-            return None
-        elif data_len > length:
-            data = data[:length]
+            data_len = len(data)
+            if data_len < length:
+                # raise ValueError('data length error: %d, %d' % (data_len, length.value))
+                return None
+            elif data_len > length:
+                data = data[:length]
+        # get string
         desc = data.rstrip(b'\0').decode('utf-8')
         return cls(data=data, description=desc)
 
