@@ -7,7 +7,7 @@
 # ==============================================================================
 # MIT License
 #
-# Copyright (c) 2019 Albert Moky
+# Copyright (c) 2020 Albert Moky
 #
 # Permission is hereby granted, free of charge, to any person obtaining a copy
 # of this software and associated documentation files (the "Software"), to deal
@@ -43,32 +43,20 @@ from typing import Union, Optional
 from udp import Hub
 from udp.hub import Cargo
 
-from .protocol import Package, Header
-from .attributes import Attribute, ChangeRequestValue
-
-
-class Info(dict):
-    """
-        1. Parameters from request
-        2. Result info from response
-    """
-
-    def __init__(self):
-        super().__init__()
-        # Bind Request
-        self.change_request: ChangeRequestValue = None
-        # Bind Response
-        self.mapped_address: (str, int) = None
-        self.changed_address: (str, int) = None
-        self.source_address: (str, int) = None
+from .protocol import Package
+from .attributes import Attribute
 
 
 class Node(ABC):
 
-    def __init__(self, host: str, port: int):
+    def __init__(self, port: int, host: str='0.0.0.0', hub: Hub=None):
         super().__init__()
         self.__local_address = (host, port)
-        self.__hub: Hub = None
+        if hub is None:
+            hub = Hub()
+            hub.open(host=host, port=port)
+            # hub.start()
+        self.__hub = hub
 
     @property
     def source_address(self) -> tuple:
@@ -88,28 +76,15 @@ class Node(ABC):
 
     @property
     def hub(self) -> Hub:
-        if self.__hub is None:
-            self.__hub = self._create_hub()
         return self.__hub
 
-    def _create_hub(self) -> Hub:
-        assert isinstance(self.__local_address, tuple), 'local address error'
-        host = self.__local_address[0]
-        port = self.__local_address[1]
-        assert port > 0, 'invalid port: %d' % port
-        hub = Hub()
-        hub.open(host=host, port=port)
-        # hub.start()
-        return hub
-
     def start(self):
-        if not self.hub.running:
-            self.hub.start()
+        # start hub
+        self.hub.start()
 
     def stop(self):
         # stop hub
-        if self.__hub is not None:
-            self.__hub.stop()
+        self.hub.stop()
 
     # noinspection PyMethodMayBeStatic
     def info(self, msg: str):
@@ -143,37 +118,36 @@ class Node(ABC):
             return None
 
     @abstractmethod
-    def parse_attribute(self, attribute: Attribute, context: dict, result: Info) -> Info:
+    def parse_attribute(self, attribute: Attribute, context: dict) -> bool:
         """
         Parse attribute
 
         :param attribute:
         :param context:
-        :param result:
-        :return:
+        :return: False on failed
         """
         raise NotImplemented
 
-    def parse_data(self, data: bytes, context: dict) -> (Optional[Header], Optional[Info]):
+    def parse_data(self, data: bytes, context: dict) -> bool:
         """
         Parse package data
 
-        :param data:
-        :param context:
-        :return: package head and results from body
+        :param data:    data package received
+        :param context: return with package head and results from body
+        :return: False on failed
         """
         # 1. parse STUN package
         pack = Package.parse(data=data)
         if pack is None:
             self.info('failed to parse package data: %d' % len(data))
-            return None, None
+            return False
         # 2. parse attributes
-        result = Info()
         attributes = Attribute.parse_all(data=pack.body)
         for item in attributes:
             # 3. process attribute
-            result = self.parse_attribute(attribute=item, context=context, result=result)
-        return pack.head, result
+            self.parse_attribute(attribute=item, context=context)
+        context['head'] = pack.head
+        return True
 
 
 def get_local_ip(remote_host: str = '8.8.8.8', remote_port: int = 80) -> Optional[str]:
